/**
  *  \file client/cscreen.cc
  *  \brief Control Screens, and Related Things
  *
  *  This contains the main classes for control screens.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/chartusr.h"
#include "client/cscreen.h"
#include "client/dlg-list.h"
#include "client/playerscr.h"
#include "client/screens.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/skin.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "client/wscanchart.h"
#include "client/history.h"
#include "client/dlg-visual.h"
#include "int/intglobal.h"
#include "client/dlg-comment.h"
#include "client/dialogs/searchdlg.h"
#include "client/dialogs/consoledlg.h"
#include "int/simple.h"
#include "client/dialogs/keymapdlg.h"
#include "game/global.h"
#include "ui/prefix.h"

//#define BLUESKIN
#ifndef BLUESKIN
#  define ICOLOR (COLOR_SHIELD+3)
#  define IFILL GRAY50_FILL
#else
#  define ICOLOR (COLOR_SHIELD+1)
#  define IFILL SOLID_FILL
#  define FCOLOR (COLOR_SHIELD+5)
#endif

WControlScreen* WControlScreen::current_cscreen;

/************************** WControlScreen::Skin ***************************/

/** \class WControlScreen::Skin
    \brief Skin for control screens */

class WControlScreen::Skin : public UISkin {
    const UIColorSet& getColors() const;
    void  outText(GfxCanvas& can,  int f, int x, int y, string_t text);
    void  outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text);
    void  drawBackground(GfxCanvas& can, GfxRect r);
};

const UIColorSet&
WControlScreen::Skin::getColors() const
{
    return DARK_COLOR_SET;
}

#ifdef BLUESKIN
void
WControlScreen::Skin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    if (isSolid()) {
        int width = fonts[f]->getTextWidth(text);
        drawBackground(can, GfxRect(x - width*parm.just_x/2, y, width, font_heights[f]));
    }
    parm.bgcolor = GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outText(can, parm, x, y, text);
}

void
WControlScreen::Skin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    if (isSolid())
        drawBackground(can, GfxRect(x - lim*parm.just_x/2, y, lim, font_heights[f]));
    parm.bgcolor = GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outTextF(can, parm, x, y, lim, text);
}

void
WControlScreen::Skin::drawBackground(GfxCanvas& can, GfxRect r)
{
    can.drawBar(r, standard_colors[ICOLOR], standard_colors[0], IFILL, GFX_ALPHA_OPAQUE);
}
#else
void
WControlScreen::Skin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    parm.bgcolor = isSolid() ? standard_colors[COLOR_DARK] : GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outText(can, parm, x, y, text);
}

void
WControlScreen::Skin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    parm.bgcolor = isSolid() ? standard_colors[COLOR_DARK] : GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outTextF(can, parm, x, y, lim, text);
}

void
WControlScreen::Skin::drawBackground(GfxCanvas& can, GfxRect r)
{
    drawSolidBar(can, r, standard_colors[COLOR_DARK]);
}
#endif

/***************************** WControlScreen ****************************/

/*
  Layout:

    ,-----------------------------------.
    |              ,------------------. |   <-- borders are 5 pixels
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              `------------------' |  <-- tile_new_y  \ 3 lines for
    |              [                  ] |  ___ chart_y2    / scanner widget
    |                                   |
    |                                   |    8 lines free here for tiles
    |                                   |
    |                                   |
    `-----------------------------------'  ---

    |  320 pixels  |
                chart_x1
 */

/** Constructor.
    \param type object type we're displaying. Used for iteration.
    \param screen_id this screen's identifier.

    Checklist for derived constructors:
    - create all tiles and add them with addTile(). You should include a
      WControlScreenHeaderTile.
    - create starchart widget and hand it to addChart() */
WControlScreen::WControlScreen(GObjectType& type, WScreenId::ScreenId screen_id)
    : screen_id(screen_id),
      type(type),
      conn_pos_changed(type.sig_pos_changed.add(this, &WControlScreen::onPosChanged)),
      scanner(),
      chart(0),
      return_value()
{
    setSkin(holder.add(new Skin()));
    setExtent(GfxRect(*ui_screen), true);
    setGeometry();
    setState(st_Modal, true);

    const int scanner_height = 2*font_heights[FONT_NORMAL];
    add(scanner, GfxRect(chart_x1+2,                chart_y2 + 5,
                         extent.w - chart_x1 - 5-2, scanner_height));
    add(holder.add(new UIPrefixArg()), GfxRect(0, 0, 0, 0));

    pushHistoryScreenId(WScreenId(screen_id, type.getCurrent()));

    current_cscreen = this;
}

WControlScreen::~WControlScreen()
{
    if (current_cscreen == this)
        current_cscreen = 0;
}

/** Compute geometry parameters. */
void
WControlScreen::setGeometry()
{
    tile_width = 315;           // FIXME: should probably be related to font size somehow
    tile_new_y = extent.h - 8*font_heights[FONT_NORMAL];
    tile_y     = 5;
    tile_x     = 5;
    chart_y2   = tile_new_y - 3*font_heights[FONT_NORMAL];
    chart_x1   = tile_width + tile_x + 5;
}

/** Selection-change handler. This is called when the user picks a
    different object on this object's type. */
void
WControlScreen::onPosChanged(int /*id*/)
{
    pushHistoryScreenId(WScreenId(screen_id, type.getCurrent()));
    update();
}

/** Add starchart widget. Must be called once. */
void
WControlScreen::addChart(WScannerChartWidget& chart)
{
    this->chart = &chart;
    add(chart, GfxRect(chart_x1, 5, extent.w - chart_x1 - 5, chart_y2 - 5));
    scanner.setChart(chart);
}

/** Add a tile, and give it a title. */
void
WControlScreen::addTile(WTile& tile, const char* title)
{
    tile.setHeading(title);
    tile.setDisplayState(WTile::sNormal);
    addTile(tile);
}

/** Add a tile. */
void
WControlScreen::addTile(WTile& tile)
{
    UIWidget::LayoutInfo info;
    tile.getLayoutInfo(info);

    if (tile_y + info.pref_size.y > extent.h - 5) {
        /* tile does not fit in this column. Would it fit when we
           remove the heading? */
        if (tile.getDisplayState() == WTile::sNormal) {
            tile.setDisplayState(WTile::sContentOnly);
            tile.getLayoutInfo(info);
            if (tile_y + info.pref_size.y <= extent.h - 5)
                goto success;
            tile.setDisplayState(WTile::sNormal);
            tile.getLayoutInfo(info);
        }

        /* start new column */
        tile_y = tile_new_y;
        tile_x += tile_width;

        /* does it fit there? */
        if (tile_y + info.pref_size.y > extent.h - 5) {
            /* nope. Remove heading if there, but we cannot do
               anything else so let it be. */
            if (tile.getDisplayState() == WTile::sNormal) {
                tile.setDisplayState(WTile::sContentOnly);
                tile.getLayoutInfo(info);
            }
        }
    }
 success:
    GfxRect tile_size(tile_x, tile_y, tile_width, info.pref_size.y);
    tile.setExtent(tile_size, true);
    tile_y += info.pref_size.y + 2;

    add(tile, tile_size);
}

/** Define keymap for this screen. */
void
WControlScreen::setKeymap(IntKeymapRef keymap)
{
    this->keymap = keymap;
}

void
WControlScreen::drawContent(GfxCanvas& can)
{
    // FIXME: optimize!
    getSkin().startDraw();
    getSkin().drawBackground(can, extent);

#ifdef BLUESKIN
    getSkin().getParameters().color = standard_colors[FCOLOR];
    drawRectangle(can, getSkin(), GfxRect(extent.x, extent.y+1, extent.w, extent.h-2));
    drawRectangle(can, getSkin(), GfxRect(extent.x+1, extent.y, extent.w-2, extent.h));

    drawRectangle(can, getSkin(), GfxRect(chart_x1-1, 5-2,        extent.w - chart_x1 - 3, chart_y2 - 5+4));
    drawRectangle(can, getSkin(), GfxRect(chart_x1-2, 5-1,        extent.w - chart_x1 - 1, chart_y2 - 5+2));

    drawRectangle(can, getSkin(), GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
#else
    drawFrameUp(can, extent);
    drawFrameDown(can, GfxRect(chart_x1-1, 5-1,        extent.w - chart_x1 - 3, chart_y2 - 5+2));
    drawFrameDown(can, GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
#endif
    can.update(extent);
}

bool
WControlScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        /* Check user-defined keymaps */
        if (keymap != 0) {
            uint32_t command = keymap->lookupCommand(e.key.code);
            if (isAtom(command)) {
                IntTemporaryContextLock lock;
                setScriptContext();
                executeStatement(atomStr(command), e.key.arg);
                return true;
            }
        }
        /* Check hardwired commands */
        switch(e.key.code) {
         case SDLK_ESCAPE:
            stop(0);
            return true;
         case SDLK_F1:
            doScreen(WScreenId::ShipScreen, *this);
            return true;
         case SDLK_F1 + ss_Shift:
         case SDLK_F1 + ss_Ctrl:
            if (chart->isValid()) {
                const GPoint pos = ((e.key.code & ss_Ctrl)
                                    ? chart->getScannerPosition()
                                    : chart->getCenterPosition());
                const int start_ship = (screen_id==1 || screen_id==11) ? type.getCurrent() : 0;
                if (getDisplayedTurn().getCurrentUniverse().getNumberOfPlayedShipsAt(pos) > 0) {
                    GObjectId id = doListShips(pos,
                                               ls_EarlyExit,
                                               start_ship,
                                               _("Visual Scanner"),
                                               _("Go to"),
                                               0 /* FIXME */);
                    if (id.isValid())
                        stopWithScreenId(WScreenId(WScreenId::ShipScreen, id.getId()));
                }
            }
            return true;
         case SDLK_F2:
            doScreen(WScreenId::PlanetScreen, *this);
            return true;
         case SDLK_F2 + ss_Shift:
            stopWithScreenId(WScreenId(WScreenId::PlanetScreen,
                                       getDisplayedTurn().getCurrentUniverse().getPlanetAt(chart->getCenterPosition(), true)));
            return true;
         case SDLK_F2 + ss_Ctrl:
            stopWithScreenId(WScreenId(WScreenId::PlanetScreen,
                                       scanner.getScannedPlanetId(false)));
            return true;
         case SDLK_F3:
            doScreen(WScreenId::BaseScreen, *this);
            return true;
         case SDLK_F3 + ss_Shift:
            stopWithScreenId(WScreenId(WScreenId::BaseScreen,
                                       getDisplayedTurn().getCurrentUniverse().getPlanetAt(chart->getCenterPosition(), true)));
            return true;
         case SDLK_F3 + ss_Ctrl:
         case SDLK_F8 + ss_Ctrl:
            stopWithScreenId(WScreenId(WScreenId::BaseScreen,
                                       scanner.getScannedPlanetId(false)));
            return true;
         case SDLK_F4:
         case SDLK_F4 + ss_Shift:
            /* Go to starchart. If we don't have a correct current
               position, we go to the starchart anyway, using the last
               position in that case. */
            if (chart->isValid()) {
                setStarchartObject(type.getObject(type.getCurrent()));
                stopWithScreenId(WScreenId(chart->getCenterPosition()));
            } else {
                stopWithScreenId(WScreenId(getStarchartPosition()));
            }
            return true;
         case SDLK_F4 + ss_Ctrl:
            if (chart->isValid()) {
                stopWithScreenId(chart->getScannerPosition());
            }
            return true;
         case SDLK_F5:
         case SDLK_F5 + ss_Shift:
            if (int pid = getDisplayedTurn().getCurrentUniverse().getPlanetAt(chart->getCenterPosition()))
                stopWithScreenId(doPlanetScan(pid));
            return true;
         case SDLK_F5 + ss_Ctrl:
            if (int pid = scanner.getScannedPlanetId(false))
                stopWithScreenId(doPlanetScan(pid));
            return true;
         case SDLK_F7:
            stopWithScreenId(doSearch());
            return true;
         case ss_Ctrl + 'h':
            stopWithScreenId(doSearchHere(chart->getCenterPosition()));
            return true;
         case SDLK_F9 + ss_Ctrl:
            if (int pid = scanner.getScannedPlanetId(false))
                doEditComment(getDisplayedTurn().getCurrentUniverse().getPlanet(pid),
                              getPlanetProperties(pid),
                              _("Enter new comment for this planet:"),
                              _("Edit Comment"));
            return true;
         case 'l':
            if (chart->isValid()) {
                GObjectId id = doListShips(chart->getScannerPosition(),
                                           ls_IncludeForeignShips,
                                           0,
                                           _("Visual Scanner"),
                                           _("Go to"),
                                           0 /* FIXME */);
                if (id.isValid())
                    stopWithScreenId(WScreenId(WScreenId::ShipScreen, id.getId()));
            }
            return true;
         case 'L':
            if (chart->isValid()) {
                GObjectId id = doListShips(chart->getCenterPosition(),
                                           ls_IncludeForeignShips,
                                           0,
                                           _("Visual Scanner"),
                                           _("Go to"),
                                           0 /* FIXME */);
                if (id.isValid())
                    stopWithScreenId(WScreenId(WScreenId::ShipScreen, id.getId()));
            }
            return true;
         case 'y':
            if (chart->isValid())
                chart->doKeyboardMode();
            return true;
         case SDLK_BACKSPACE + ss_Alt:
            stopWithScreenId(doScreenHistoryDialog(true));
            return true;
         case SDLK_BACKSPACE:
            stopWithScreenId(popHistoryScreenId(true));
            return true;
         case ss_Alt + 'c':
            {
                IntTemporaryContextLock lock;
                setScriptContext();
                doConsole();
            }
            return true;
         case ss_Alt + 'k':
            doKeymapDebugger(keymap);
            return true;
        }
    }
    return UIFixedWidget::handleEvent(e, second_pass);
}

/** Go to object #id. This one is tolerant in that it silently ignores invalid Ids. */
void
WControlScreen::setCurrent(int id)
{
    if (type.isExistant(id))
        type.setCurrent(id);
}

/** Browse through objects. */
void
WControlScreen::doBrowse(bool forward, bool marked)
{
    int n;
    if (forward)
        n = type.getNextWrap(type.getCurrent(), marked);
    else
        n = type.getPreviousWrap(type.getCurrent(), marked);
    setCurrent(n);
}

/** Get scanner position. */
GPoint
WControlScreen::getScannerPosition() const
{
    return chart->getScannerPosition();
}

/** Run this control screen.
    \return Id of screen to activate (1..3 for ship/planet/base, 0 for race screen, etc.) */
WScreenId
WControlScreen::runScreen()
{
    update();
    ui_root->add(*this);
    UIFixedWidget::run();
    ui_root->close(*this);
    return return_value;
}

/** Exit this control screen, and go to specified screen.
    \param id Screen id. It may be invalid; in this case, nothing happens. */
void
WControlScreen::stopWithScreenId(WScreenId id)
{
    if (id.isValid()) {
        return_value = id;
        stop(0);
    }
}
