/**
  *  \file client/playerscr.cc
  *  \brief Player Main Screen
  *
  *  Main screen, user interface and multiplexing to other control screens.
  *  PCC 1.x source: cc.pas, scrmux.pas.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <SDL.h>
#include "client/screens.h"
#include "client/chartusr.h"
#include "client/cscreen.h"
#include "client/dlg-list.h"
#include "client/playerscr.h"
#include "client/vcrplay.h"
#include "game/global.h"
#include "game/objlist.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/font.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/ui.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "cpluslib/holder.h"
#include "client/dlg-mbox.h"
#include "cpluslib/format.h"
#include "client/scr-sim.h"
#include "game/sim-game.h"
#include "client/history.h"
#include "game/stat.h"
#include "client/scr-score.h"
#include "game/team.h"
#include "game/racename.h"
#include "client/dialogs/teamedit.h"
#include "client/dialogs/commandedit.h"
#include "ui/widgets/keytranslator.h"
#include "client/dialogs/consoledlg.h"
#include "client/dialogs/searchdlg.h"
#include "resmgr/resmgr.h"

class WPixmapButton : public UIButton {
    Ptr<GfxPixmap> pix;
 public:
    WPixmapButton(int aid, string_t aname, UIKey akey, Ptr<GfxPixmap> apix);
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
};

WPixmapButton::WPixmapButton(int aid, string_t aname,
                             UIKey akey, Ptr<GfxPixmap> apix)
    : UIButton(aid, aname, akey), pix(apix)
{ }

void
WPixmapButton::drawContent(GfxCanvas& can)
{
    GfxParameters parm;
    int text_height = fonts[FONT_SMALL]->getTextHeight(getName());
    if (pix && extent.h > text_height-2)
        pix->blit(can, extent.x + 1, extent.y + 1,
                  GfxRect(0, 0, extent.w - 2, extent.h - text_height - 2));
    parm.color = standard_colors[COLOR_WHITE];
    parm.bgcolor = standard_colors[COLOR_BLACK];
    parm.font = fonts[FONT_SMALL];
    outTextF(can, parm,
             extent.x + 1, extent.y + extent.h - text_height - 1,
             extent.w - 2, getName());
    drawFrameDown(can, extent);
    can.update(extent);
}

void
WPixmapButton::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint pt (fonts[FONT_SMALL]->getTextWidth(getName()) + 2,
                 fonts[FONT_SMALL]->getTextHeight(getName()) + 2);
    if (pix) {
        pt.x = pix->getWidth() + 2;
        pt.y += pix->getHeight();
    }
    info.flags = 0;
    info.min_size = info.pref_size = pt;
}

/***************************** WPlayerScreen *****************************/

class WPlayerScreen : public UIGroup {
    Ptr<GfxPixmap> pix;
    static UIKeyTranslator::Item keys[];
    ObjectHolder holder;
    UIHBoxLayout layout;
    UIVBoxLayout vlayout;
 public:
    WPlayerScreen();
    void drawContent(GfxCanvas& can);
    bool handleEvent(const UIEvent& e, bool second_pass);
    bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
};

UIKeyTranslator::Item
WPlayerScreen::keys[] = {
    { SDLK_ESCAPE, cm_Escape, false }
};

WPlayerScreen::WPlayerScreen()
    : UIGroup(UINullLayout::instance /* will be overridden */), layout(5, 5), vlayout(5, 5)
{
    setLayout(layout);
    setExtent(GfxRect(*ui_screen));
    pix = requestPixmap("bgtile");
    UIGroup& g = holder.add(new UIGroup(vlayout));
    g.add(holder.add(new WPixmapButton(1, _("F1 - Starships"), SDLK_F1,
                             requestPixmap("menu.ship"))).addCommand());
    g.add(holder.add(new WPixmapButton(2, _("F2 - Planets"), SDLK_F2,
                             requestPixmap("menu.planet"))).addCommand());
    g.add(holder.add(new WPixmapButton(3, _("F3 - Starbases"), SDLK_F3,
                             requestPixmap("menu.base"))).addCommand());
    g.add(holder.add(new WPixmapButton(4, _("F4 - Starchart"), SDLK_F4,
                             requestPixmap("menu.chart"))).addCommand());
    g.add(holder.add(new UISpacer()));
    add(g);
    g.pack();

    UIGroup& fgroup = holder.add(new UIGroup(holder.add(new UIFlowLayout(4, true))));
    fgroup.add(holder.add(new UIButton(_("T - Teams"), 't')));
    fgroup.add(holder.add(new UIButton(_("S - Scores"), 's')));
    fgroup.add(holder.add(new UIButton(_("B - Battle Simulator"), 'b')));
    fgroup.add(holder.add(new UIButton(_("M - Messages"), 'm')));
    fgroup.add(holder.add(new UIButton(_("W - Write Message"), 'w')));
    fgroup.add(holder.add(new UIButton(_("V - Combat Recorder"), 'v')));
    fgroup.add(holder.add(new UIButton(_("ESC - Exit"), SDLK_ESCAPE)));

    UIGroup& vgroup = holder.add(new UIGroup(vlayout));
    vgroup.add(holder.add(new UISpacer()));
    vgroup.add(fgroup);
    add(vgroup);

    doLayout();

    add(holder.add(new UIKeyTranslator(keys, countof(keys))));
}

void
WPlayerScreen::drawContent(GfxCanvas& can)
{
    drawTiledArea(can, GfxRect(extent.x + 1, extent.y + 1,
                               extent.w - 2, extent.h - 2),
                  pix, standard_colors[COLOR_DARK], 16);
    drawFrameUp(can, extent);

    GfxParameters parm;
    parm.font  = fonts[FONT_TITLE];
    parm.color = standard_colors[COLOR_WHITE];

    outText(can, parm, 200,  20, player_racenames.getFullName(getPlayerId()));
    outText(can, parm, 200,  60, format(_("Turn %d"), getDisplayedTurn().getTurnNumber()));
    outText(can, parm, 200,  80, format(_("%d message%!1{s%}"), getDisplayedTurn().getInbox().getCount()));
    outText(can, parm, 200, 100, format(_("%d combat recording%!1{s%}"), getDisplayedTurn().getVcrs().getCount()));

    can.update(extent);
}

/** Play VCRs. */
static void
playVcrs(GVcrDatabase& db)
{
    // HACK HACK HACK
    GClassicVcrDatabase* cdb = dynamic_cast<GClassicVcrDatabase*>(&db);
    if (!cdb)
        return;

    ObjectHolder holder;
    WVcrSelector sel(*cdb);
    WVcrSelectorWindow w(sel);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(holder.add(new UIButton(cm_VCR_Prev, "\036", '-')).addCommand());
    but.add(holder.add(new UIButton(cm_VCR_Next, "\037", '+')).addCommand());
    but.add(holder.add(new UIButton(cm_Enter, _("ENTER - Play"), SDLK_RETURN)).addCommand());
    but.add(holder.add(new UISpacer()));
    but.add(holder.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE)).addStop());

    w.setLayout(UIVBoxLayout::instance5);
    w.add(sel);
    w.add(but);
    w.pack();
    ui_root->centerWidget(w);
    ui_root->add(w);
    ui_root->run();
    ui_root->close(w);
}

bool
WPlayerScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch(cmd.code) {
     case cm_Escape:
        stop(-1);
        return true;
     case 1:
     case 2:
     case 3:
     case 6:
     case 10:
        doMainLoop(doSelectObjectForScreen(WScreenId::ScreenId(cmd.code)));
        return true;
     case 4:
        doMainLoop(getStarchartPosition());
        return true;
     default:
        return UIGroup::handleCommand(cmd, sender);
    }
}

#include "ui/scroll.h"
static void
testScrollWindow()
{
    ObjectHolder h;
    UIGroup g(UIVBoxLayout::instance5);
    for (int i = 0; i < 100; ++i)
        g.add(h.add(new UIButton(i, format("%d", i), ' ')).addStop());

    UIScrollableGroup sgroup(g);
    UIWindow win("Test", UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(sgroup);
    win.add(h.add(new UIButton(999, "OK", SDLK_RETURN)).addStop());
    win.pack();

    ui_root->centerWidget(win);
    ui_root->runModal(win);
}

bool
WPlayerScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        if (e.key.code == 'v') {
            if (getDisplayedTurn().getVcrs().getCount())
                playVcrs(getDisplayedTurn().getVcrs());
            return true;
        } else if (e.key.code == 'm') {
            doInboxScreen();
            return true;
        } else if (e.key.code == 'w') {
            doSendOrReviseMessage();
            return true;
        } else if (e.key.code == 'b') {
            // FIXME: this should be converted to use WScreenId as well.
            // This is not trivial as GSimulatorRealGameInterface is a 'game'
            // class and does not know about the WScreenId 'client' class.
            // A possible method would be to implement an own descendant of
            // GSimulatorRealGameInterface which overrides selectGameShip().
            GSimulatorRealGameInterface rgi(getDisplayedTurn().getCurrentUniverse());
            doMainLoop(WScreenId(WScreenId::ScreenId(doSimulator(rgi)), 0));
            return true;
        } else if (e.key.code == 's') {
            if (isStatisticsFileLoaded() && getStatistics().getNumTurns() != 0)
                doScoreScreen(getStatistics(), getPlayerId());
            return true;
        } else if (e.key.code == 'x') {
            testScrollWindow();
            return true;
        } else if (e.key.code == 't') {
            editTeams();
            return true;
        } else if (e.key.code == ss_Alt + 'e') {
            editCommands();
            return true;
        } else if (e.key.code == ss_Alt + 'c') {
            doConsole();
            return true;
        } else if (e.key.code == SDLK_F7) {
            doMainLoop(doSearch());
            return true;
        } else if (e.key.code == SDLK_BACKSPACE) {
            doMainLoop(popHistoryScreenId(false));
            return true;
        } else if (e.key.code == SDLK_BACKSPACE + ss_Alt) {
            doMainLoop(doScreenHistoryDialog(false));
            return true;
        }
    }
    return UIGroup::handleEvent(e, second_pass);
}

/***************************** Main Routines *****************************/

int
doPlayerScreen()
{
    WPlayerScreen ps;
    ui_root->add(ps);
    int rv = ui_root->run();
    ui_root->close(ps);
    return rv;
}

/** Main loop. Giant state machine. (cc.pas::RunScreen in PCC 1.x) */
int
doMainLoop(WScreenId state)
{
    while (1) {
        switch (state.getScreenId()) {
         case WScreenId::ShipScreen:
            if (state.getId())
                global_ship_type.setCurrent(state.getId());
            state = doShipScreen();
            break;

         case WScreenId::PlanetScreen:
            if (state.getId())
                global_planet_type.setCurrent(state.getId());
            state = doPlanetScreen();
            break;

         case WScreenId::BaseScreen:
            if (state.getId())
                global_base_type.setCurrent(state.getId());
            state = doBaseScreen();
            break;
            
         case WScreenId::Starchart:
            setStarchartPosition(state.getPos());
            state = doStarchart();
            break;

         default:
            return state.getScreenId();
        }
    }
}

/** Get type associated with a screen number. Screen numbers are function
    key numbers for the appropriate screens, without formal \#defines. */
GObjectType&
getObjectType(int screen, GGameTurn& trn)
{
    switch (screen) {
     case 1: case 11:
        return global_ship_type;
     case 2: case 12:
        return global_planet_type;
     case 3: case 13:
        return global_base_type;
        // FIXME: fleets
     default:
        FAIL();
    }
}

/** Get type associated with a screen number. */
GObjectType&
getObjectType(int screen)
{
    return getObjectType(screen, getDisplayedTurn());
}

/** Do object selection window for a particular screen Id.
    \returns chosen Id or 0 */
int
doSelectObject(int screen)
{
    switch (screen) {
     case 1: case 11:
        return chooseShip();
     case 2: case 12:
        return choosePlanet();
     case 3: case 13:
        return chooseBase();
        // FIXME: ship history, fleets
     default:
        return 0;
    }
}

/** Do a control screen. Does selection window and calls stopWithScreenId
    if needed. */
void
doScreen(WScreenId::ScreenId screen_id, WControlScreen& me)
{
    if (int i = doSelectObject(screen_id))
        me.stopWithScreenId(WScreenId(screen_id, i));
}
