/**
  *  \file client/search.cc
  *  \brief Search Function Main Code
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/search.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "int/bytecode.h"
#include "int/exec.h"
#include "int/expr.h"
#include "int/if/ionif.h"
#include "int/if/mineif.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "int/if/ufoif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "int/lex.h"
#include "int/parseexpr.h"
#include "int/unary.h"
#include "ui/msgbox.h"
#include "util/misc.h"
#include "util/nls.h"


/** Prepare a search query. Parses a TUserSearchQuery into a TParsedSearchQuery.
    On error, displays a dialog box, and returns false. On success, returns true.
    \param query [out] Parsed query
    \param in    [in] User query */
bool
prepareSearchQuery(TParsedSearchQuery& query, const TUserSearchQuery& in)
{
    /* Prepare object types */
    query.search_obj = in.search_obj;

    /* Prepare query string */
    string_t expr = strTrim(in.query);
    if (expr.size() == 0) {
        /* Make a "match any" query */
        query.match = query.MatchAny;
        return true;
    }

    switch (in.search_kind) {
     case TUserSearchQuery::MatchName:
        query.name = strUCase(expr);
        if (!stringValue(expr, query.id))
            if (!(expr[0] == '#' && stringValue(expr.substr(1), query.id)))
                query.id = -1;
        query.match = query.MatchName;
        return true;

     case TUserSearchQuery::MatchTrue:
     case TUserSearchQuery::MatchFalse:
        try {
            // Parse expression
            IntTokenizer tok(expr);
            std::auto_ptr<IntExprNode> node(IntExpressionParser(tok).parse());
            if (tok.getCurrentToken() != tok.tEnd)
                throw IntError::garbageAtEnd(true);

            // Compile expression
            query.expr_bco = new IntBytecodeObject();
            node->compileValue(*query.expr_bco, IntCompilationContext());

            // Negate if necessary
            if (in.search_kind == TUserSearchQuery::MatchFalse)
                query.expr_bco->addInstruction(IntOpcode::maUnary, unNot2, 0);
            query.expr_bco->relocate();

            query.match = query.MatchExpr;
        }
        catch (IntError& e) {
            messageBox(e.what(), _("Search Object"));
            return false;
        }
        return true;

     case TUserSearchQuery::MatchLocation:
        if (!query.location.parseCoordinates(expr)) {
            messageBox(_("Please enter a valid X,Y coordinate pair."), _("Search Object"));
            return false;
        }
        query.match = query.MatchLocation;
        return true;

     default:
        return false;
    }
}

/*************************** executeSearchQuery **************************/

/** Search planets.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here
    \param mustHaveBase [in] true to restrict search on starbases only */
static void
searchPlanets(const TParsedSearchQuery& query, GObjectList& result, TSearchResultStatus& status, bool mustHaveBase)
{
    for (int pid = global_anyplanet_type.getNextIndex(0); pid != 0; pid = global_anyplanet_type.getNextIndex(pid)) {
        bool match;
        if (mustHaveBase && !global_anyplanet_type.getObject(pid).hasBase())
            continue;
        if (!global_anyplanet_type.getObject(pid).hasAnyPlanetData())
            continue;

        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = pid == query.id
                || strUCase(global_planet_type.getObject(pid).getName()).find(query.name) != string_t::npos
                || strUCase(IntValue::toStringOf(getPlanetProperties(pid)[pp_Comment], false)).find(query.name) != string_t::npos;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = global_anyplanet_type.getObject(pid).getPos() == query.location;
            break;
         case TParsedSearchQuery::MatchExpr:
            try {
                IntExecutionContext exec;
                exec.pushNewContext(new IntPlanetContext(pid));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                exec.run();
                match = getBoolValue(exec.getResult()) > 0;
                status.did_evaluate = true;
            }
            catch (IntError& e) {
                status.had_error = true;
                status.last_error = e.what();
                match = false;
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            result.addObject(global_anyplanet_type, pid);
        }
    }
}

/** Search starships.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here */
static void
searchShips(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    for (int sid = global_anyship_type.getNextIndex(0); sid != 0; sid = global_anyship_type.getNextIndex(sid)) {
        bool match;
        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = sid == query.id
                || strUCase(global_anyship_type.getObject(sid).getName()).find(query.name) != string_t::npos
                || strUCase(IntValue::toStringOf(getShipProperties(sid)[sp_Comment], false)).find(query.name) != string_t::npos;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = global_anyship_type.getObject(sid).getPos() == query.location;
            break;
         case TParsedSearchQuery::MatchExpr:
            try {
                IntExecutionContext exec;
                exec.pushNewContext(new IntShipContext(sid));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                exec.run();
                match = getBoolValue(exec.getResult()) > 0;
                status.did_evaluate = true;
            }
            catch (IntError& e) {
                status.had_error = true;
                status.last_error = e.what();
                match = false;
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            result.addObject(global_anyship_type, sid);
        }
    }
}

/** Search Ufos.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here */
static void
searchUfos(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    GUfoContainer& con = getDisplayedTurn().getCurrentUniverse().getUfos();
    for (GUfoContainer::iterator it = con.begin(); it != con.end(); ++it) {
        bool match;
        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = it->getId() == query.id || strUCase(it->getName()).find(query.name) != string_t::npos;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = distanceSquared(it->getPos(), query.location) <= square(it->getRadius());
            break;
         case TParsedSearchQuery::MatchExpr:
            try {
                IntExecutionContext exec;
                exec.pushNewContext(new IntUfoContext(con, it));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                exec.run();
                match = getBoolValue(exec.getResult()) > 0;
                status.did_evaluate = true;
            }
            catch (IntError& e) {
                status.had_error = true;
                status.last_error = e.what();
                match = false;
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            // FIXME: this sucks.
            result.addObject(global_ufo_type, 1 + std::distance(con.begin(), it));
        }
    }
}

/** Search minefields.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here */
static void
searchMinefields(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    for (int mid = global_minefield_type.getNextIndex(0); mid != 0; mid = global_minefield_type.getNextIndex(mid)) {
        bool match;
        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = mid == query.id;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = distanceSquared(global_minefield_type.getObject(mid).getPos(), query.location) <= global_minefield_type.getObject(mid).getUnits();
            break;
         case TParsedSearchQuery::MatchExpr:
            try {
                IntExecutionContext exec;
                exec.pushNewContext(new IntMinefieldContext(mid));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                exec.run();
                match = getBoolValue(exec.getResult()) > 0;
                status.did_evaluate = true;
            }
            catch (IntError& e) {
                status.had_error = true;
                status.last_error = e.what();
                match = false;
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            result.addObject(global_minefield_type, mid);
        }
    }
}

/** Search ion storms.
    \param query  [in] Search query
    \param result [out] Append result here
    \param status [out] Update status here */
static void
searchIonStorms(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    for (int iid = global_ion_type.getNextIndex(0); iid != 0; iid = global_ion_type.getNextIndex(iid)) {
        bool match;
        status.had_object = true;
        switch (query.match) {
         case TParsedSearchQuery::MatchAny:
            match = true;
            break;
         case TParsedSearchQuery::MatchName:
            match = (iid == query.id) || strUCase(global_ion_type.getObject(iid).getName()).find(query.name) != string_t::npos;
            break;
         case TParsedSearchQuery::MatchLocation:
            match = distanceSquared(global_ion_type.getObject(iid).getPos(), query.location) <= square(global_ion_type.getObject(iid).getRadius());
            break;
         case TParsedSearchQuery::MatchExpr:
            try {
                IntExecutionContext exec;
                exec.pushNewContext(new IntIonContext(iid));
                exec.pushNewFrame(new IntExecutionFrame(query.expr_bco));
                exec.run();
                match = getBoolValue(exec.getResult()) > 0;
                status.did_evaluate = true;
            }
            catch (IntError& e) {
                status.had_error = true;
                status.last_error = e.what();
                match = false;
            }
            break;
         default:
            match = false;
            break;
        }

        if (match) {
            result.addObject(global_ion_type, iid);
        }
    }
}

/** Execute search query.
    \param query  [in] Parsed query
    \param status [out] Result status
    \param result [out] Result */
void
executeSearchQuery(const TParsedSearchQuery& query, TSearchResultStatus& status, GObjectList& result)
{
    // Search planets and bases
    if (query.search_obj & TUserSearchQuery::SearchPlanets) {
        searchPlanets(query, result, status, false);
    } else if (query.search_obj & TUserSearchQuery::SearchBases) {
        searchPlanets(query, result, status, true);
    } else {
        // Don't search planets
    }

    // Search ships
    if (query.search_obj & TUserSearchQuery::SearchShips) {
        searchShips(query, status, result);
    }

    // Search Ufos
    if (query.search_obj & TUserSearchQuery::SearchUfos) {
        searchUfos(query, status, result);
    }

    // Search rest
    if (query.search_obj & TUserSearchQuery::SearchOthers) {
        searchMinefields(query, status, result);
        searchIonStorms(query, status, result);
    }
}
