/**
  *  \file flakconf.cc
  *  \brief FLAK Configuration
  *
  *  This contains the FLAK configuration structure which is needed in
  *  addition to the normal host configuration (PConfig). This module
  *  contains the parser which reads a user-supplied config file into
  *  our structure.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "flak/flakconf.h"
#include "util/confpars.h"
#include "util/nls.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/misc.h"

/** Global FLAK configuration object. */
FlakConfig flak_config;

namespace {
    /// FLAK Configuration Parser.
    class FlakConfigParser : public ConfigParser {
     public:
        void assignNumber(int line_id, string_t value, int& out);
        void assign(int line_id, string_t name, string_t value, string_t line);
        void error(int line_id, string_t msg);
    };
}

void
FlakConfigParser::assign(int line_id, string_t name, string_t value, string_t /*line*/)
{
    if (strCaseCmp(name, "ratingbeamscale") == 0)
        assignNumber(line_id, value, flak_config.RatingBeamScale);
    else if (strCaseCmp(name, "ratingtorpscale") == 0)
        assignNumber(line_id, value, flak_config.RatingTorpScale);
    else if (strCaseCmp(name, "ratingbayscale") == 0)
        assignNumber(line_id, value, flak_config.RatingBayScale);
    else if (strCaseCmp(name, "ratingmassscale") == 0)
        assignNumber(line_id, value, flak_config.RatingMassScale);
    else if (strCaseCmp(name, "ratingpebonus") == 0)
        assignNumber(line_id, value, flak_config.RatingPEBonus);
    else if (strCaseCmp(name, "RatingFullAttackBonus") == 0)
        assignNumber(line_id, value, flak_config.RatingFullAttackBonus);
    else if (strCaseCmp(name, "RatingRandomBonus") == 0)
        assignNumber(line_id, value, flak_config.RatingRandomBonus);
    else if (strCaseCmp(name, "StartingDistanceShip") == 0)
        assignNumber(line_id, value, flak_config.StartingDistanceShip);
    else if (strCaseCmp(name, "StartingDistancePlanet") == 0)
        assignNumber(line_id, value, flak_config.StartingDistancePlanet);
    else if (strCaseCmp(name, "StartingDistancePerPlayer") == 0)
        assignNumber(line_id, value, flak_config.StartingDistancePerPlayer);
    else if (strCaseCmp(name, "StartingDistancePerFleet") == 0)
        assignNumber(line_id, value, flak_config.StartingDistancePerFleet);
    else if (strCaseCmp(name, "CompensationShipScale") == 0)
        assignNumber(line_id, value, flak_config.CompensationShipScale);
    else if (strCaseCmp(name, "CompensationBeamScale") == 0)
        assignNumber(line_id, value, flak_config.CompensationBeamScale);
    else if (strCaseCmp(name, "CompensationTorpScale") == 0)
        assignNumber(line_id, value, flak_config.CompensationTorpScale);
    else if (strCaseCmp(name, "CompensationFighterScale") == 0)
        assignNumber(line_id, value, flak_config.CompensationFighterScale);
    else if (strCaseCmp(name, "CompensationLimit") == 0)
        assignNumber(line_id, value, flak_config.CompensationLimit);
    else if (strCaseCmp(name, "CompensationMass100KTScale") == 0)
        assignNumber(line_id, value, flak_config.CompensationMass100KTScale);
    else if (strCaseCmp(name, "CompensationAdjust") == 0)
        assignNumber(line_id, value, flak_config.CompensationAdjust);
    else if (strCaseCmp(name, "CyborgDebrisRate") == 0)
        assignNumber(line_id, value, flak_config.CyborgDebrisRate);
    else if (strCaseCmp(name, "MaximumFleetSize") == 0)
        assignNumber(line_id, value, flak_config.MaximumFleetSize);
    else if (strCaseCmp(name, "SendUtilData") == 0) {
        if (stringMatch("Yes", value))
            flak_config.SendUtildata = true;
        else if (stringMatch("No", value))
            flak_config.SendUtildata = false;
        else
            error(line_id, _("Invalid boolean setting"));
    } else
        error(line_id, _("Invalid keyword"));
}

void
FlakConfigParser::assignNumber(int line_id, string_t value, int& out)
{
    if (!stringValue(value, out))
        error(line_id, _("Invalid number"));
}

void
FlakConfigParser::error(int line_id, string_t msg)
{
    console.write(LOG_ERROR, format(_("line %d: %s")) << line_id << msg);
}

/** Initialize FLAK Configuration to defaults. */
void
initFlakConfig()
{
    flak_config.RatingBeamScale            = 1;
    flak_config.RatingTorpScale            = 1;
    flak_config.RatingBayScale             = 8;
    flak_config.RatingMassScale            = 1;
    flak_config.RatingPEBonus              = 20;
    flak_config.RatingFullAttackBonus      = 20;
    flak_config.RatingRandomBonus          = 20;
    flak_config.StartingDistanceShip       = 26000;
    flak_config.StartingDistancePlanet     = 10000;
    flak_config.StartingDistancePerPlayer  = 1000;
    flak_config.StartingDistancePerFleet   = 5000;
    flak_config.CompensationShipScale      = 0;
    flak_config.CompensationBeamScale      = 30;
    flak_config.CompensationTorpScale      = 90;
    flak_config.CompensationFighterScale   = 90;
    flak_config.CompensationLimit          = 500;
    flak_config.CompensationMass100KTScale = 0;
    flak_config.CompensationAdjust         = 0;
    flak_config.SendUtildata               = true;
    flak_config.CyborgDebrisRate           = 75;
    flak_config.MaximumFleetSize           = 999;
}

/** Parse FLAK configuration file.
    \param file        input file
    \param in_section  if true, assume we are inside a FLAK configuration
                       block. If false, look for a "%flak" section delimiter
                       first. */
void
loadFlakConfig(Stream& file, bool in_section)
{
    FlakConfigParser fp;
    fp.configure("flak", in_section);
    fp.parseFile(file);
}
