/**
  *  \file int/if/guiif.cc
  *  \brief Interpreter Game Interface - GUI Components
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "int/if/guiif.h"
#include "client/actions.h"
#include "client/cscreen.h"
#include "client/dlg-ammo.h"
#include "client/dlg-build.h"
#include "client/dlg-fcode.h"
#include "client/dlg-mission.h"
#include "client/dlg-structure.h"
#include "client/dlg-supplies.h"
#include "client/dlg-tax.h"
#include "client/dlg-tech.h"
#include "client/screens.h"
#include "client/widgets/specview.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/host.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/shipquery.h"
#include "int/args.h"
#include "int/compound.h"
#include "int/exec.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "util/modinit.h"
#include "util/nls.h"
#include "util/strlist.h"
#include "ui/event.h"

void
flushUI()
{
    // Tell widgets to order their redraws
    if (haveDisplayedTurn())
        getDisplayedTurn().getCurrentUniverse().doScreenUpdates();

    // Perform redraws
    if (ui_root)
        ui_root->performDeferredRedraws();
}

/** Global command "MessageBox". */
static void
IFMessageBox(IntExecutionContext& exc, IntArgBlock& args)
{
    /* MessageBox <text>[, <heading>] */

    // Validate arguments
    args.checkArgumentCount(1, 2);

    string_t content;
    string_t heading = _("Message");
    if (!checkStringArg(content, args.getNext()))
        return;
    checkStringArg(heading, args.getNext());

    // Do it
    flushUI();
    if (exc.checkForBreak())
        return;
    messageBox(content, heading);
}

/** Global command "UI.Update". */
static void
IFUIUpdate(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    /* UI.Update [<forceFullRedrawFlag>] */

    args.checkArgumentCount(0, 1);

    // FIXME: this is not 100% according to rules, because it treats a missing argument
    // different from a null argument
    int mode = 0;
    if (args.getNumArgs() > 0)
        mode = getBoolValue(args.getNext());

    if (mode >= 0) {
        // Regular redraw
        flushUI();

        // Force full redraw
        if (mode > 0 && ui_root)
            ui_root->drawWidget(true);
    }
}

/** Global command "UI.Input". */
static void
IFUIInput(IntExecutionContext& exc, IntArgBlock& args)
{
    /* UI.Input <prompt>[, <title>, <maxChars>, <flags>, <default>] */
    args.checkArgumentCount(1, 5);

    string_t prompt;
    string_t title;
    string_t defaultText;
    int32_t  maxChars = 255;
    int32_t  flags = 0;
    int32_t  width = ui_root->getExtent().w / 2;

    // Mandatory argument
    if (!checkStringArg(prompt, args.getNext()))
        return;

    // Optional arguments
    title = prompt;
    checkStringArg(title, args.getNext());
    checkIntArg(maxChars, args.getNext(), 0, 32000);
    checkFlagArg(flags, &width, args.getNext(), "NHPFGM");
    checkStringArg(defaultText, args.getNext());

    // Convert flags
    //   N = numeric
    //   H = on high ASCII
    //   P = password masking
    //   F = frame
    //   G = game charset
    //   M = width is in ems
    uint32_t uiFlags = 0;
    if (flags & 1)
        uiFlags |= UIInputLine::ilf_NumbersOnly;
    if (flags & 2)
        uiFlags |= UIInputLine::ilf_NoHi;
    if (flags & 4)
        uiFlags |= UIInputLine::ilf_Hidden;
    if (flags & 8)
        uiFlags |= UIInputLine::ilf_Framed;
    if (flags & 16)
        uiFlags |= UIInputLine::ilf_GameChars;

    if ((flags & 32) == 0 && font_ems[FONT_TITLE] > 0) {
        // width given in in pixels. Convert to ems.
        width /= font_ems[FONT_TITLE];
    }

    // Build a UIInputLine object
    flushUI();
    if (exc.checkForBreak())
        return;
    if (UIInputLine::run(defaultText, maxChars, title, prompt, width, uiFlags)) {
        IntStringValue sv(defaultText);
        exc.setVariable("UI.RESULT", &sv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

static void
IFUIInputNumber(IntExecutionContext& exc, IntArgBlock& args)
{
    // Check arguments
    args.checkArgumentCount(1, 5);

    string_t title;
    if (!checkStringArg(title, args.getNext()))
        return;

    int32_t min = 0;
    checkIntArg(min, args.getNext());

    int32_t max = 10000;
    checkIntArg(max, args.getNext());

    int32_t current = 0;
    checkIntArg(current, args.getNext());

    /* The fifth argument is the help screen Id, which we currently do not use. */

    // Adjust arguments
    if (max < min)
        std::swap(min, max);

    // Do it.
    flushUI();
    if (exc.checkForBreak())
        return;
    UIDecimalSelector nsel(0, min, max, 10);
    nsel.setValue(current);  /* value will be clipped into valid range */
    if (nsel.doStandardDialog(title, title, 0)) {
        IntIntValue iv(nsel.getValue(), false);
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

static void
IFUIChooseObject(IntExecutionContext& exc, IntArgBlock& args)
{
    // Check arguments
    int32_t screen;
    args.checkArgumentCount(1);
    if (!checkIntArg(screen, args.getNext(), 0, 100))
        return;
    if (!WScreenId::isValidScreenId(screen))
        throw IntError::rangeError();

    // Do it
    flushUI();
    if (exc.checkForBreak())
        return;
    WScreenId result = doSelectObjectForScreen(WScreenId::ScreenId(screen));
    if (result.isValid()) {
        IntIntValue iv(result.getId(), false);
        exc.setVariable("UI.RESULT", &iv);
    } else {
        exc.setVariable("UI.RESULT", 0);
    }
}

static void
IFUIMessage(IntExecutionContext& exc, IntArgBlock& args)
{
    using std::tolower;

    args.checkArgumentCount(1, 3);

    // Read arguments
    string_t text;
    string_t title   = _("Message");
    string_t buttons = _("OK");
    if (!checkStringArg(text, args.getNext()))
        return;
    checkStringArg(title, args.getNext());
    checkStringArg(buttons, args.getNext());

    buttons = strTrim(buttons);

    // Prepare dialog
    UIMessageBox msgbox(text, title);
    int id = 0;
    do {
        string_t b = strFirst(buttons, " ");
        if (!b.empty()) {
            ++id;
            msgbox.addButton(id, b, tolower(uint8_t(b[0])));
        }
    } while (strRemove(buttons, " "));
    msgbox.addKey(1, SDLK_RETURN);
    msgbox.addKey(id, SDLK_ESCAPE);

    // Do it
    if (id == 0) {
        exc.setVariable("UI.RESULT", 0);
    } else {
        flushUI();
        if (exc.checkForBreak())
            return;
        int result = msgbox.run();
        IntIntValue iv(result, false);
        exc.setVariable("UI.RESULT", &iv);
    }
}

/** Internal global command "CC$CSBrowse". */
static void
IFCCCSBrowse(IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    /* CC$CSBrowse <forwardFlag>, <onlyMarkedFlag> */
    args.checkArgumentCount(2);

    WControlScreen* cs = WControlScreen::getCurrentControlScreen();
    if (!cs)
        throw IntError::contextError();

    bool forward, marked;
    if (checkBoolArg(forward, args.getNext()) && checkBoolArg(marked, args.getNext()))
        cs->doBrowse(forward, marked);
}

static void
IFCCChangeFCode(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        // Planet fcode change
        GFCodeList list(GFCodeList::getInstance(), *pl);
        WFCodeWindow dialog(_("Change Friendly Code"), list);
        dialog.setFCode(pl->getFCode());
        ui_root->centerWidget(dialog);
        ui_root->add(dialog);
        if (dialog.run() == cm_Enter)
            pl->setFCode(dialog.getFCode());
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        // Ship fcode change
        GFCodeList list(GFCodeList::getInstance(), *sh);
        WFCodeWindow dialog(_("Change Friendly Code"), list);
        dialog.setFCode(sh->getFCode());
        ui_root->centerWidget(dialog);
        ui_root->add(dialog);
        if (dialog.run() == cm_Enter)
            sh->setFCode(dialog.getFCode());
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangePE(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        // Ship PE change
        StringList d;
        d.add(0, _("0 - none"));
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            d.add(i, format("%X - %s", i, player_racenames.getShortName(i)));

        UIStandardListbox lb(0, -1, 240, d);
        lb.scrollToKey(sh->getPE());
        if (lb.doStandardListbox(_("Change Primary Enemy"), 0))
            sh->setPE(lb.getKey());
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeMission(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        doShipMissionChange(*sh);
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        if (pl->hasBase()) {
            doBaseChangeMission(*pl);
        } else {
            throw IntError::contextError();
        }
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildStructures(IntExecutionContext& exc, IntArgBlock& args)
{
    int32_t screen = 0;
    args.checkArgumentCount(1);
    checkIntArg(screen, args.getNext(), 0, 2);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        doPlanetBuildScreen(*pl, screen);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeTech(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed() && pl->hasBase()) {
        doTechUpgradeDialog(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCChangeTaxes(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed()) {
        doTaxation(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCSellSupplies(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed()) {
        WSellSuppliesDialog::doDialog(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuySupplies(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed()) {
        WUndoSellSuppliesDialog::doDialog(getDisplayedTurn() /* FIXME */, pl->getId());
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCGive(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GObject* obj = exc.getCurrentObject();
    if (!obj || !obj->isExistant() || !obj->isPlayed()) {
        throw IntError::contextError();
    } else if (!host.isPHost() || !config.CPEnableGive()) {
        // ignore
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(obj)) {
        // give planet
        doGivePlanet(*pl, getDisplayedTurn().getCommands(pl->getOwner()));
    } else if (GShip* sh = dynamic_cast<GShip*>(obj)) {
        // give ship
        doGiveShip(*sh, getDisplayedTurn().getCommands(sh->getOwner()));
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCShipSpec(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GShip* sh = dynamic_cast<GShip*>(exc.getCurrentObject());
    if (sh != 0 && sh->isExistant()) {
        GShipQuery q;
        q.initForExistingShip(getDisplayedTurn().getCurrentUniverse(), sh->getId());
        WSpecView sv;
        sv.setQuery(q);
        sv.doStandardDialog();
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildAmmo(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed() && pl->hasBase()) {
        doBaseBuildAmmo(*pl);
    } else {
        throw IntError::contextError();
    }
}

static void
IFCCBuildShip(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    flushUI();
    if (exc.checkForBreak())
        return;
    GPlanet* pl = dynamic_cast<GPlanet*>(exc.getCurrentObject());
    if (pl != 0 && pl->isExistant() && pl->isPlayed() && pl->hasBase()) {
        WShipBuildScreenWithActor bs(*pl);
        ui_root->centerWidget(bs);
        ui_root->add(bs);
        ui_root->run();
        ui_root->close(bs);
    } else {
        throw IntError::contextError();
    }
}

static bool
checkBreak()
{
    bool result = checkBreakEvent();
    if (result) {
        result = UIMessageBox(_("PCC2 is currently executing a script. "
                                "Do you want to abort that script?"),
                              _("Script Interpreter")).doYesNoDialog();
        ui_root->performDeferredRedraws();
    }
    return result;
}

static void
initInterpreterGuiInterface()
{
    // Values
    defineGlobalValue("SYSTEM.GUI", makeBoolValue(1));

    // Procedures
    defineGlobalProcedure("CC$BUILDAMMO",       IFCCBuildAmmo);
    defineGlobalProcedure("CC$BUILDSHIP",       IFCCBuildShip);
    defineGlobalProcedure("CC$BUILDSTRUCTURES", IFCCBuildStructures);
    defineGlobalProcedure("CC$BUYSUPPLIES",     IFCCBuySupplies);
    defineGlobalProcedure("CC$CHANGEFCODE",     IFCCChangeFCode);
    defineGlobalProcedure("CC$CHANGEMISSION",   IFCCChangeMission);
    defineGlobalProcedure("CC$CHANGEPE",        IFCCChangePE);
    defineGlobalProcedure("CC$CHANGETAXES",     IFCCChangeTaxes);
    defineGlobalProcedure("CC$CHANGETECH",      IFCCChangeTech);
    defineGlobalProcedure("CC$CSBROWSE",        IFCCCSBrowse);
    defineGlobalProcedure("CC$GIVE",            IFCCGive);
    defineGlobalProcedure("CC$SELLSUPPLIES",    IFCCSellSupplies);
    defineGlobalProcedure("CC$SHIPSPEC",        IFCCShipSpec);
    defineGlobalProcedure("MESSAGEBOX",         IFMessageBox);
    defineGlobalProcedure("UI.CHOOSEOBJECT",    IFUIChooseObject);
    defineGlobalProcedure("UI.INPUT",           IFUIInput);
    defineGlobalProcedure("UI.INPUTNUMBER",     IFUIInputNumber);
    defineGlobalProcedure("UI.MESSAGE",         IFUIMessage);
    defineGlobalProcedure("UI.UPDATE",          IFUIUpdate);

    IntExecutionContext::setBreakHandler(checkBreak);
}

// Constraints: after mod_intgame_common, after mod_int, with mod_base_gui
const ModuleInitializer mod_intgame_gui[] = {
    { OnStartup, initInterpreterGuiInterface },
    { 0, 0 }
};
