/**
  *  \file int/if/specif.cc
  *  \brief Interpreter Game Interface: Specification Access
  *
  *  This implements access to Torpedo, Beam, and Engine specifications.
  *
  *  \todo It's pretty ugly that we have three almost identical classes here.
  *  Can we do better?
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/specif.h"
#include "int/value.h"
#include "game/spec.h"
#include "game/host.h"
#include "game/config.h"
#include "int/if/ifutil.h"
#include "cpluslib/misc.h"
#include "int/interror.h"
#include "int/data.h"
#include "int/compound.h"
#include "int/args.h"

/**********************                             **********************/
/********************** Part I: Interface Functions **********************/
/**********************                             **********************/

/** Implementation of 'Torpedo(Id)'. */
IntValue*
IFTorpedoGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_TORPS))
        return 0;
    return IntSimpleContextValue::createFromNewContext(new IntTorpedoContext(id));
}

/** Implementation of 'ForEach Torpedo'. */
IntContext*
IFTorpedoMake()
{
    return new IntTorpedoContext(1);
}

/** Implementation of 'Launcher(Id)'. */
IntValue*
IFLauncherGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_TORPS))
        return 0;
    return IntSimpleContextValue::createFromNewContext(new IntLauncherContext(id));
}

/** Implementation of 'ForEach Launcher'. */
IntContext*
IFLauncherMake()
{
    return new IntLauncherContext(1);
}

/** Implementation of 'Beam(Id)'. */
IntValue*
IFBeamGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_BEAMS))
        return 0;
    return IntSimpleContextValue::createFromNewContext(new IntBeamContext(id));
}

/** Implementation of 'ForEach Beam'. */
IntContext*
IFBeamMake()
{
    return new IntBeamContext(1);
}

/** Implementation of 'Engine(Id)'. */
IntValue*
IFEngineGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!checkIntArg(id, args.getNext(), 1, NUM_ENGINES))
        return 0;
    return IntSimpleContextValue::createFromNewContext(new IntEngineContext(id));
}

/** Implementation of 'ForEach Engine'. */
IntContext*
IFEngineMake()
{
    return new IntEngineContext(1);
}

/***************************                   ***************************/
/*************************** Part II: Contexts ***************************/
/***************************                   ***************************/

namespace {
    enum { ComponentProperty, WeaponProperty, EngineProperty };
}

// Mapping for torpedoes and launchers:
static const TIntMapping torpedo_map[] = {
    { "COST.D",        icpCostD,     ComponentProperty },
    { "COST.M",        icpCostM,     ComponentProperty },
    { "COST.MC",       icpCostMC,    ComponentProperty },
    { "COST.STR",      icpCostStr,   ComponentProperty },
    { "COST.T",        icpCostT,     ComponentProperty },
    { "DAMAGE",        iwpDamage,    WeaponProperty },
    { "ID",            icpId,        ComponentProperty },
    { "KILL",          iwpKill,      WeaponProperty },
    { "MASS",          icpMass,      ComponentProperty },
    { "NAME",          icpName,      ComponentProperty },
    { "NAME.SHORT",    icpNameShort, ComponentProperty },
    { "TECH",          icpTech,      ComponentProperty },
    { "TECH.TORPEDO",  icpTech,      ComponentProperty },
};

// Mapping for beams:
static const TIntMapping beam_map[] = {
    { "COST.D",        icpCostD,     ComponentProperty },
    { "COST.M",        icpCostM,     ComponentProperty },
    { "COST.MC",       icpCostMC,    ComponentProperty },
    { "COST.STR",      icpCostStr,   ComponentProperty },
    { "COST.T",        icpCostT,     ComponentProperty },
    { "DAMAGE",        iwpDamage,    WeaponProperty },
    { "ID",            icpId,        ComponentProperty },
    { "KILL",          iwpKill,      WeaponProperty },
    { "MASS",          icpMass,      ComponentProperty },
    { "NAME",          icpName,      ComponentProperty },
    { "NAME.SHORT",    icpNameShort, ComponentProperty },
    { "TECH",          icpTech,      ComponentProperty },
    { "TECH.BEAM",     icpTech,      ComponentProperty },
};

// Mapping for engines:
static const TIntMapping engine_map[] = {
    { "COST.D",        icpCostD,         ComponentProperty },
    { "COST.M",        icpCostM,         ComponentProperty },
    { "COST.MC",       icpCostMC,        ComponentProperty },
    { "COST.STR",      icpCostStr,       ComponentProperty },
    { "COST.T",        icpCostT,         ComponentProperty },
    { "FUELFACTOR",    iepFuelFactor,    EngineProperty },
    { "ID",            icpId,            ComponentProperty },
    { "NAME",          icpName,          ComponentProperty },
    { "NAME.SHORT",    icpNameShort,     ComponentProperty },
    { "SPEED$",        iepEfficientWarp, EngineProperty },
    { "TECH",          icpTech,          ComponentProperty },
    { "TECH.ENGINE",   icpTech,          ComponentProperty },
};

/*************************** IntTorpedoContext ***************************/

IntTorpedoContext::IntTorpedoContext(int id)
    : id(id)
{ }

IntTorpedoContext::~IntTorpedoContext()
{ }

int32_t
IntTorpedoContext::lookup(string_t name)
{
    return lookupName(name, torpedo_map, countof(torpedo_map));
}

void
IntTorpedoContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntTorpedoContext::get(int32_t index)
{
    switch (torpedo_map[index].domain) {
     case ComponentProperty: return getComponentProperty(getTorp(id).getTorpedo(), IntComponentProperty(torpedo_map[index].index));
     case WeaponProperty:    return getWeaponProperty   (getTorp(id).getTorpedo(), IntWeaponProperty   (torpedo_map[index].index), true);
    }
    return 0;
}

bool
IntTorpedoContext::next()
{
    if (id < NUM_TORPS) {
        ++id;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntTorpedoContext::clone()
{
    return new IntTorpedoContext(id);
}

GObject*
IntTorpedoContext::getObject()
{
    return 0;
}

/*************************** IntLauncherContext **************************/

IntLauncherContext::IntLauncherContext(int id)
    : id(id)
{ }

IntLauncherContext::~IntLauncherContext()
{ }

int32_t
IntLauncherContext::lookup(string_t name)
{
    return lookupName(name, torpedo_map, countof(torpedo_map));
}

void
IntLauncherContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntLauncherContext::get(int32_t index)
{
    switch (torpedo_map[index].domain) {
     case ComponentProperty: return getComponentProperty(getTorp(id), IntComponentProperty(torpedo_map[index].index));
     case WeaponProperty:    return getWeaponProperty   (getTorp(id), IntWeaponProperty   (torpedo_map[index].index), true);
    }
    return 0;
}

bool
IntLauncherContext::next()
{
    if (id < NUM_TORPS) {
        ++id;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntLauncherContext::clone()
{
    return new IntLauncherContext(id);
}

GObject*
IntLauncherContext::getObject()
{
    return 0;
}

/***************************** IntBeamContext ****************************/

IntBeamContext::IntBeamContext(int id)
    : id(id)
{ }

IntBeamContext::~IntBeamContext()
{ }

int32_t
IntBeamContext::lookup(string_t name)
{
    return lookupName(name, beam_map, countof(beam_map));
}

void
IntBeamContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntBeamContext::get(int32_t index)
{
    switch (beam_map[index].domain) {
     case ComponentProperty: return getComponentProperty(getBeam(id), IntComponentProperty(beam_map[index].index));
     case WeaponProperty:    return getWeaponProperty   (getBeam(id), IntWeaponProperty   (beam_map[index].index), false);
    }
    return 0;
}

bool
IntBeamContext::next()
{
    if (id < NUM_BEAMS) {
        ++id;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntBeamContext::clone()
{
    return new IntBeamContext(id);
}

GObject*
IntBeamContext::getObject()
{
    return 0;
}

/************************** IntEngineFuelFactor **************************/

namespace {
    /** Implementation of Engine().FuelFactor(). */
    class IntEngineFuelFactor : public IntIndexableValue {
        const GEngine& e;
     public:
        IntEngineFuelFactor(const GEngine& e)
            : e(e)
            { }
        ~IntEngineFuelFactor()
            { }
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& /*args*/, IntValue* /*value*/)
            { throw IntError::notAssignable(); }
        IntContext* makeFirstContext()
            { throw IntError::typeError(IntError::tyIterable); }
        string_t toString(bool /*readable*/)
            { return "#<array>"; }
        void store(IntTagNode&, Stream&)
            { throw IntError::notSerializable(); }
        IntValue* clone()
            { return new IntEngineFuelFactor(e); }
    };
}

IntValue*
IntEngineFuelFactor::get(IntArgBlock& args)
{
    int32_t warp;
    args.checkArgumentCount(1);
    if (!checkIntArg(warp, args.getNext(), 0, 9))
        return 0;
    return makeIntValue(e.getFuelFactor(warp));
}

/**************************** IntEngineContext ***************************/

IntEngineContext::IntEngineContext(int id)
    : id(id)
{ }

IntEngineContext::~IntEngineContext()
{ }

int32_t
IntEngineContext::lookup(string_t name)
{
    return lookupName(name, engine_map, countof(engine_map));
}

void
IntEngineContext::set(int32_t index, IntValue* value)
{
    switch (engine_map[index].domain) {
     case EngineProperty:
        setEngineProperty(getEngine(id), IntEngineProperty(engine_map[index].index), value);
        break;
     default:
        throw IntError::notAssignable();
    }
}

IntValue*
IntEngineContext::get(int32_t index)
{
    switch (engine_map[index].domain) {
     case ComponentProperty: return getComponentProperty(getEngine(id), IntComponentProperty(engine_map[index].index));
     case EngineProperty:    return getEngineProperty   (getEngine(id), IntEngineProperty   (engine_map[index].index));
    }
    return 0;
}

bool
IntEngineContext::next()
{
    if (id < NUM_ENGINES) {
        ++id;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntEngineContext::clone()
{
    return new IntEngineContext(id);
}

GObject*
IntEngineContext::getObject()
{
    return 0;
}

/***********************                           ***********************/
/*********************** Part III: Property Access ***********************/
/***********************                           ***********************/

/** Get property of a component.
    \param comp Component to query
    \param icp  Property to query */
IntValue*
getComponentProperty(const GComponent& comp, IntComponentProperty icp)
{
    switch (icp) {
     case icpMass:
        return makeIntValue(comp.getMass());
     case icpTech:
        return makeIntValue(comp.getTech());
     case icpCostT:
        return makeIntValue(comp.getCost().get(el_Tritanium));
     case icpCostD:
        return makeIntValue(comp.getCost().get(el_Duranium));
     case icpCostM:
        return makeIntValue(comp.getCost().get(el_Molybdenum));
     case icpCostMC:
        return makeIntValue(comp.getCost().get(el_Money));
     case icpCostSup:
        return makeIntValue(comp.getCost().get(el_Supplies));
     case icpCostStr:
        return makeStringValue(comp.getCost().toCargospec());
     case icpName:
        return makeStringValue(comp.getName());
     case icpNameShort:
        return makeStringValue(comp.getShortName());
     case icpId:
        return makeIntValue(comp.getId());
    }
    return 0;
}

/** Get property of a weapon. */
IntValue*
getWeaponProperty(const GWeapon& w, IntWeaponProperty iwp, bool isTorpedo)
{
    /* We're doubling torpedo bang/kill values when appropriate */
    int factor;
    if (isTorpedo && (host.isTHost() || !config.AllowAlternativeCombat()))
        factor = 2;
    else
        factor = 1;

    switch (iwp) {
     case iwpKill:
        return makeIntValue(factor * w.getKill());
     case iwpDamage:
        return makeIntValue(factor * w.getDamage());
    }
    return 0;
}

/** Get property of an engine. */
IntValue*
getEngineProperty(const GEngine& e, IntEngineProperty iep)
{
    switch (iep) {
     case iepEfficientWarp:
        return makeIntValue(e.getMaxEfficientWarp());
     case iepFuelFactor:
        return new IntEngineFuelFactor(e);
    }
    return 0;
}

/** Set property of an engine. */
void
setEngineProperty(GEngine& e, IntEngineProperty iep, IntValue* value)
{
    int32_t val;
    switch (iep) {
     case iepEfficientWarp:
        if (checkIntArg(val, value, 0, 9))
            e.setMaxEfficientWarp(val);
        break;
     default:
        throw IntError::notAssignable();
    }
}
